/*
 * Copyright (C) 2014 Rolf Eike Beer <eike@sf-mail.de>
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "hydwidget.h"

#include "osmnode.h"

#include <marble/GeoDataDocument.h>
#include <marble/GeoDataPlacemark.h>
#include <marble/GeoDataTreeModel.h>
#include <marble/MarbleModel.h>
#include <QBuffer>
#include <QDebug>
#include <QNetworkReply>
#include <QXmlStreamReader>

const QString HydWidget::emg = QLatin1String("emergency");
const QString HydWidget::amen = QLatin1String("amenity");
const QString HydWidget::fh = QLatin1String("fire_hydrant");
const QString HydWidget::wtank = QLatin1String("water_tank");
static const QString hw = QLatin1String("highway");
static const QString eap = QLatin1String("emergency_access_point");

using namespace Marble;

HydWidget::HydWidget(QWidget *parent)
	: Marble::MarbleWidget(parent),
	m_netmgr(new QNetworkAccessManager(this))
{
	// Load the OpenStreetMap map
	setMapThemeId(QLatin1String("earth/openstreetmap/openstreetmap.dgml"));
	setShowOverviewMap(false);
	setShowCompass(false);

	connect(m_netmgr, SIGNAL(finished(QNetworkReply*)),
			SLOT(slotDownloadDone(QNetworkReply*)));
}

HydWidget::~HydWidget()
{
}

void HydWidget::readData(QIODevice *r)
{
	QXmlStreamReader xs(r);
	GeoDataDocument *document = new GeoDataDocument();

	while (!xs.atEnd()) {
		if (xs.readNextStartElement()) {
			static const QString coord = QLatin1String("coordinates");
			static const QString edta = QLatin1String("ExtendedData");
			static const QString dta = QLatin1String("Data");
			static const QString nd = QLatin1String("node");

			if (xs.name() != nd)
				continue;

			QStringList descr;
			double x = 0.0, y = 0.0;
			qint64 id;
			bool b = false;

			if (xs.attributes().hasAttribute(QLatin1String("id"))) {
				const QString idStr = xs.attributes().value(QLatin1String("id")).toString();

				descr << QLatin1String("<a href=\"http://www.openstreetmap.org/node/") + idStr + "\">id: " + idStr + "</a>";
				id = idStr.toLongLong(&b);
			}
			if (!b) {
				qDebug() << "invalid or missing id";
				continue;
			}

			if (m_hyds.contains(id))
				continue;

			if (xs.attributes().hasAttribute(QLatin1String("lat"))) {
				const QStringRef lStr = xs.attributes().value(QLatin1String("lat"));

				bool b;
				y = lStr.toString().toDouble(&b);
				if (!b)
					continue;
			} else {
				continue;
			}

			if (xs.attributes().hasAttribute(QLatin1String("lon"))) {
				const QStringRef lStr = xs.attributes().value(QLatin1String("lon"));

				bool b;
				x = lStr.toString().toDouble(&b);
				if (!b)
					continue;
			} else {
				continue;
			}

			QHash<QString, QString> attributes;

			while (!xs.atEnd() && (!xs.isEndElement() || (xs.name() != nd))) {
				static const QString tag = QLatin1String("tag");
				static const QString name = QLatin1String("name");

				xs.readNextStartElement();

				if (xs.name() != tag)
					continue;

				if (!xs.attributes().hasAttribute(QLatin1String("k")) || !xs.attributes().hasAttribute(QLatin1String("v")))
					continue;

				const QString k = xs.attributes().value(QLatin1String("k")).toString();
				const QString v = xs.attributes().value(QLatin1String("v")).toString();

				attributes[k] = v;

				descr << k + QLatin1String(": ") + v;
			}

			// Now all attributes of the returned node are in the hash attributes.
			// Decide if the node is of any interest.
			QScopedPointer<OSMNode> node;

			const auto itEnd = m_filter.constEnd();
			for (auto it = m_filter.constBegin(); (it != itEnd) && node.isNull(); it++) {
				const QString v = attributes.value(it.key());
				if (v.isEmpty())
					continue;

				if (v != it.value())
					continue;

				if (v == fh)
					node.reset(new FireHydrant(x, y, attributes));
				else if (v == wtank)
					node.reset(new WaterTank(x, y, attributes));
				else if (v == eap)
					node.reset(new EmergencyAccessPoint(x, y, attributes));
				else
					node.reset(new OSMNode(x, y, attributes));
			}

			if (node.isNull())
				continue;

			GeoDataPlacemark *place = node->getPlacemark();
			place->setCoordinate(x, y, 0.0, GeoDataCoordinates::Degree);
			place->setDescription(descr.join(QLatin1String("<br/>")));

			m_hyds[id] = node.take();

			document->append( place );
		}
	}

	if (document->featureList().isEmpty()) {
		qDebug() << "data contained no points";
		delete document;
		return;
	}

	// Add the document to MarbleWidget's tree model
	model()->treeModel()->addDocument( document );
}

void HydWidget::slotDownloadDone(QNetworkReply *r)
{
	r->deleteLater();

	if (r->error() != QNetworkReply::NoError)
		qDebug() << r->errorString();
	else
		readData(r);
}

void HydWidget::loadDataForArea(qreal x0, qreal y0, qreal x1, qreal y1)
{
	/*
	 * (
	 * 	 node["amenity"="fire_hydrant"]  (52.25,9.5,52.3,9.6);
	 * 	 node["emergency"="fire_hydrant"](52.25,9.5,52.3,9.6);
	 * );
	 * out;
	 */

	const QByteArray bbox = QByteArray::number(y0) + "%2C" + QByteArray::number(x0) + "%2C" + QByteArray::number(y1) + "%2C" + QByteArray::number(x1);

	QStringList nodefilter;
	const auto itEnd = m_filter.constEnd();
	for (auto it = m_filter.constBegin(); it != itEnd; it++)
		nodefilter << QString("node%5B%22" + it.key() + "%22%3D%22" + it.value() + "%22%5D(" + bbox + ")%3B");

	const QUrl link = QUrl::fromPercentEncoding("http://overpass-api.de/api/interpreter?data=(" + nodefilter.join(QString()).toAscii() + ")%3Bout%3B");
	qDebug() << link;
	m_netmgr->get(QNetworkRequest(link));
}

void HydWidget::reload(ControlWidget::NodeTypes types)
{
	qreal la0, lo0, la1, lo1;

	m_filter.clear();
	if (types & ControlWidget::FireHydrants) {
		m_filter.insert(amen, fh);
		m_filter.insert(emg, fh);
	}
	if (types & ControlWidget::WaterTanks)
		m_filter.insert(emg, wtank);
	if (types & ControlWidget::EmergencyAccessPoints)
		m_filter.insert(hw, eap);

	if (m_filter.isEmpty())
		return;

	geoCoordinates(0, 0, la0, lo0);
	geoCoordinates(width(), height(), la1, lo1);

	const qreal dla = qAbs<qreal>(la1 - la0);
	const qreal dlo = qAbs<qreal>(lo1 - lo0);

	la0 -= (dla / 2);
	la1 += (dla / 2);
	lo0 += (dlo / 2);
	lo1 -= (dlo / 2);

	loadDataForArea(la0, lo1, la1, lo0);
}

#include "moc_hydwidget.cpp"
